@@ -1,6 +1,6 @@
 /*
  * ngIRCd -- The Next Generation IRC Daemon
- * Copyright (c)2001-2012 Alexander Barton (alex@barton.de) and Contributors.
+ * Copyright (c)2001-2013 Alexander Barton (alex@barton.de) and Contributors.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -47,10 +47,6 @@
 # include <netinet/ip.h>
 #endif
 
-#ifdef HAVE_STDINT_H
-# include <stdint.h>			/* e.g. for Mac OS X */
-#endif
-
 #ifdef TCPWRAP
 # include <tcpd.h>			/* for TCP Wrappers */
 #endif
@@ -67,6 +63,7 @@
 #include "client.h"
 #include "class.h"
 #include "conf.h"
+#include "conn-encoding.h"
 #include "conn-ssl.h"
 #include "conn-zip.h"
 #include "conn-func.h"
@@ -85,10 +82,12 @@
 #define MAX_COMMANDS_SERVER_MIN 10
 #define MAX_COMMANDS_SERVICE 10
 
+#define SD_LISTEN_FDS_START 3
+
 
 static bool Handle_Write PARAMS(( CONN_ID Idx ));
 static bool Conn_Write PARAMS(( CONN_ID Idx, char *Data, size_t Len ));
-static int New_Connection PARAMS(( int Sock ));
+static int New_Connection PARAMS(( int Sock, bool IsSSL ));
 static CONN_ID Socket2Index PARAMS(( int Sock ));
 static void Read_Request PARAMS(( CONN_ID Idx ));
 static unsigned int Handle_Buffer PARAMS(( CONN_ID Idx ));
@@ -122,6 +121,44 @@ static void cb_Read_Resolver_Result PARAMS((int sock, UNUSED short what));
 static void cb_Connect_to_Server PARAMS((int sock, UNUSED short what));
 static void cb_clientserver PARAMS((int sock, short what));
 
+time_t idle_t = 0;
+
+
+/**
+ * Get number of sockets available from systemd(8).
+ *
+ * ngIRCd needs to implement its own sd_listen_fds(3) function and can't
+ * use the one provided by systemd itself, because the sockets will be
+ * used in a forked child process with a new PID, and this would trigger
+ * an error in the standard implementation.
+ *
+ * @return Number of sockets available, -1 if sockets have already been
+ *         initialized, or 0 when no sockets have been passed.
+ */
+static int
+my_sd_listen_fds(void)
+{
+	const char *e;
+	int count;
+
+	/* Check if LISTEN_PID exists; but we ignore the result, because
+	 * normally ngircd forks a child before checking this, and therefore
+	 * the PID set in the environment is always wrong ... */
+	e = getenv("LISTEN_PID");
+	if (!e || !*e)
+		return 0;
+
+	e = getenv("LISTEN_FDS");
+	if (!e || !*e)
+		return -1;
+	count = atoi(e);
+#ifdef HAVE_UNSETENV
+	unsetenv("LISTEN_FDS");
+#endif
+
+	return count;
+}
+
 
 /**
  * IO callback for listening sockets: handle new connections. This callback
@@ -134,7 +171,7 @@ static void
 cb_listen(int sock, short irrelevant)
 {
 	(void) irrelevant;
-	(void) New_Connection(sock);
+	(void) New_Connection(sock, false);
 }
 
 
@@ -152,7 +189,7 @@ cb_listen_ssl(int sock, short irrelevant)
 	int fd;
 
 	(void) irrelevant;
-	fd = New_Connection(sock);
+	fd = New_Connection(sock, true);
 	if (fd < 0)
 		return;
 	io_event_setcb(My_Connections[fd].sock, cb_clientserver_ssl);
@@ -187,7 +224,7 @@ cb_connserver(int sock, UNUSED short what)
 	if (server < 0) {
 		Log(LOG_ERR, "Connection on socket %d to \"%s\" aborted!",
 		    sock, My_Connections[idx].host);
-		Conn_Close(idx, "Connection aborted!", NULL, false);
+		Conn_Close(idx, "Connection aborted", NULL, false);
 		return;
 	}
 
@@ -208,7 +245,7 @@ cb_connserver(int sock, UNUSED short what)
 			    My_Connections[idx].host, Conf_Server[server].port,
  			    idx, strerror(err));
 
-		Conn_Close(idx, "Can't connect!", NULL, false);
+		Conn_Close(idx, "Can't connect", NULL, false);
 
 		if (ng_ipaddr_af(&Conf_Server[server].dst_addr[0])) {
 			/* more addresses to try... */
@@ -285,7 +322,7 @@ cb_connserver_login_ssl(int sock, short unused)
 		return;
 	case -1:
 		Log(LOG_ERR, "SSL connection on socket %d failed!", sock);
-		Conn_Close(idx, "Can't connect!", NULL, false);
+		Conn_Close(idx, "Can't connect", NULL, false);
 		return;
 	}
 
@@ -332,13 +369,13 @@ cb_clientserver(int sock, short what)
 
 #ifdef SSL_SUPPORT
 /**
- * IO callback for established SSL-enabled client and server connections.
+ * IO callback for new SSL-enabled client and server connections.
  *
  * @param sock	Socket descriptor.
  * @param what	IO specification (IO_WANTREAD/IO_WANTWRITE/...).
  */
 static void
-cb_clientserver_ssl(int sock, short what)
+cb_clientserver_ssl(int sock, UNUSED short what)
 {
 	CONN_ID idx = Socket2Index(sock);
 
@@ -355,14 +392,11 @@ cb_clientserver_ssl(int sock, short what)
 	case 0:
 		return;	/* EAGAIN: callback will be invoked again by IO layer */
 	default:
-		Conn_Close(idx, "SSL accept error, closing socket", "SSL accept error", false);
+		Conn_Close(idx,
+			   "SSL accept error, closing socket", "SSL accept error",
+			   false);
 		return;
 	}
-	if (what & IO_WANTREAD)
-		Read_Request(idx);
-
-	if (what & IO_WANTWRITE)
-		Handle_Write(idx);
 
 	io_event_setcb(sock, cb_clientserver);	/* SSL handshake completed */
 }
@@ -370,14 +404,13 @@ cb_clientserver_ssl(int sock, short what)
 
 
 /**
- * Initialize connecion module.
+ * Initialize connection module.
  */
 GLOBAL void
 Conn_Init( void )
 {
 	CONN_ID i;
 
-	/* Speicher fuer Verbindungs-Pool anfordern */
 	Pool_Size = CONNECTION_POOL;
 	if ((Conf_MaxConnections > 0) &&
 		(Pool_Size > Conf_MaxConnections))
@@ -457,7 +490,7 @@ Conn_CloseAllSockets(int ExceptOf)
  * @returns		Number of listening sockets created.
  */
 static unsigned int
-ports_initlisteners(array *a, const char *listen_addr, void (*func)(int,short))
+Init_Listeners(array *a, const char *listen_addr, void (*func)(int,short))
 {
 	unsigned int created = 0;
 	size_t len;
@@ -473,8 +506,9 @@ ports_initlisteners(array *a, const char *listen_addr, void (*func)(int,short))
 			continue;
 		}
 		if (!io_event_create( fd, IO_WANTREAD, func )) {
-			Log( LOG_ERR, "io_event_create(): Could not add listening fd %d (port %u): %s!",
-						fd, (unsigned int) *port, strerror(errno));
+			Log(LOG_ERR,
+			    "io_event_create(): Can't add fd %d (port %u): %s!",
+			    fd, (unsigned int) *port, strerror(errno));
 			close(fd);
 			port++;
 			continue;
@@ -496,14 +530,76 @@ Conn_InitListeners( void )
 {
 	/* Initialize ports on which the server should accept connections */
 	unsigned int created = 0;
-	char *copy, *listen_addr;
+	char *af_str, *copy, *listen_addr;
+	int count, fd, i, addr_len;
+	ng_ipaddr_t addr;
 
 	assert(Conf_ListenAddress);
 
+	count = my_sd_listen_fds();
+	if (count < 0) {
+		Log(LOG_INFO,
+		    "Not re-initializing listening sockets of systemd(8) ...");
+		return 0;
+	}
+	if (count > 0) {
+		/* systemd(8) passed sockets to us, so don't try to initialize
+		 * listening sockets on our own but use the passed ones */
+		LogDebug("Initializing %d systemd sockets ...", count);
+		for (i = 0; i < count; i++) {
+			fd = SD_LISTEN_FDS_START + i;
+			addr_len = (int)sizeof(addr);
+			getsockname(fd, (struct sockaddr *)&addr, (socklen_t*)&addr_len);
+#ifdef WANT_IPV6
+			if (addr.sin4.sin_family != AF_INET && addr.sin4.sin_family != AF_INET6)
+#else
+			if (addr.sin4.sin_family != AF_INET)
+#endif
+			{
+				/* Socket is of unsupported type! For example, systemd passed in
+				 * an IPv6 socket but ngIRCd isn't compiled with IPv6 support. */
+				switch (addr.sin4.sin_family)
+				{
+					case AF_UNSPEC: af_str = "AF_UNSPEC"; break;
+					case AF_UNIX: af_str = "AF_UNIX"; break;
+					case AF_INET: af_str = "AF_INET"; break;
+#ifdef AF_INET6
+					case AF_INET6: af_str = "AF_INET6"; break;
+#endif
+#ifdef AF_NETLINK
+					case AF_NETLINK: af_str = "AF_NETLINK"; break;
+#endif
+					default: af_str = "unknown"; break;
+				}
+				Log(LOG_CRIT,
+				    "Socket %d is of unsupported type \"%s\" (%d), have to ignore it!",
+				    fd, af_str, addr.sin4.sin_family);
+				close(fd);
+				continue;
+			}
+
+			Init_Socket(fd);
+			if (!io_event_create(fd, IO_WANTREAD, cb_listen)) {
+				Log(LOG_ERR,
+				    "io_event_create(): Can't add fd %d: %s!",
+				    fd, strerror(errno));
+				continue;
+			}
+			Log(LOG_INFO,
+			    "Initialized socket %d from systemd(8): %s:%d.", fd,
+			    ng_ipaddr_tostr(&addr), ng_ipaddr_getport(&addr));
+			created++;
+		}
+		return created;
+	}
+
+	/* not using systemd socket activation, initialize listening sockets: */
+
 	/* can't use Conf_ListenAddress directly, see below */
 	copy = strdup(Conf_ListenAddress);
 	if (!copy) {
-		Log(LOG_CRIT, "Cannot copy %s: %s", Conf_ListenAddress, strerror(errno));
+		Log(LOG_CRIT, "Cannot copy %s: %s", Conf_ListenAddress,
+		    strerror(errno));
 		return 0;
 	}
 	listen_addr = strtok(copy, ",");
@@ -511,9 +607,11 @@ Conn_InitListeners( void )
 	while (listen_addr) {
 		ngt_TrimStr(listen_addr);
 		if (*listen_addr) {
-			created += ports_initlisteners(&Conf_ListenPorts, listen_addr, cb_listen);
+			created += Init_Listeners(&Conf_ListenPorts,
+						  listen_addr, cb_listen);
 #ifdef SSL_SUPPORT
-			created += ports_initlisteners(&Conf_SSLOptions.ListenPorts, listen_addr, cb_listen_ssl);
+			created += Init_Listeners(&Conf_SSLOptions.ListenPorts,
+						  listen_addr, cb_listen_ssl);
 #endif
 		}
 
@@ -540,7 +638,12 @@ Conn_ExitListeners( void )
 	int *fd;
 	size_t arraylen;
 
+	/* Get number of listening sockets to shut down. There can be none
+	 * if ngIRCd has been "socket activated" by systemd. */
 	arraylen = array_length(&My_Listeners, sizeof (int));
+	if (arraylen < 1)
+		return;
+
 	Log(LOG_INFO,
 	    "Shutting down all listening sockets (%d total) ...", arraylen);
 	fd = array_start(&My_Listeners);
@@ -571,7 +674,7 @@ InitSinaddrListenAddr(ng_ipaddr_t *addr, const char *listen_addrstr, UINT16 Port
 	ret = ng_ipaddr_init(addr, listen_addrstr, Port);
 	if (!ret) {
 		assert(listen_addrstr);
-		Log(LOG_CRIT, "Can't bind to [%s]:%u: can't convert ip address \"%s\"",
+		Log(LOG_CRIT, "Can't bind to [%s]:%u: can't convert ip address \"%s\"!",
 						listen_addrstr, Port, listen_addrstr);
 	}
 	return ret;
@@ -623,8 +726,9 @@ NewListener(const char *listen_addr, UINT16 Port)
 
 	af = ng_ipaddr_af(&addr);
 	sock = socket(af, SOCK_STREAM, 0);
-	if( sock < 0 ) {
-		Log(LOG_CRIT, "Can't create socket (af %d) : %s!", af, strerror(errno));
+	if (sock < 0) {
+		Log(LOG_CRIT, "Can't create socket (af %d) : %s!", af,
+		    strerror(errno));
 		return -1;
 	}
 
@@ -634,22 +738,23 @@ NewListener(const char *listen_addr, UINT16 Port)
 		return -1;
 
 	if (bind(sock, (struct sockaddr *)&addr, ng_ipaddr_salen(&addr)) != 0) {
-		Log(LOG_CRIT, "Can't bind socket to address %s:%d - %s",
-			ng_ipaddr_tostr(&addr), Port, strerror(errno));
+		Log(LOG_CRIT, "Can't bind socket to address %s:%d - %s!",
+		    ng_ipaddr_tostr(&addr), Port, strerror(errno));
 		close(sock);
 		return -1;
 	}
 
-	if( listen( sock, 10 ) != 0 ) {
-		Log( LOG_CRIT, "Can't listen on socket: %s!", strerror( errno ));
-		close( sock );
+	if (listen(sock, 10) != 0) {
+		Log(LOG_CRIT, "Can't listen on socket: %s!", strerror(errno));
+		close(sock);
 		return -1;
 	}
 
 	/* keep fd in list so we can close it when ngircd restarts/shuts down */
-	if (!array_catb( &My_Listeners,(char*) &sock, sizeof(int) )) {
-		Log( LOG_CRIT, "Can't add socket to My_Listeners array: %s!", strerror( errno ));
-		close( sock );
+	if (!array_catb(&My_Listeners, (char *)&sock, sizeof(int))) {
+		Log(LOG_CRIT, "Can't add socket to My_Listeners array: %s!",
+		    strerror(errno));
+		close(sock);
 		return -1;
 	}
 
@@ -749,8 +854,7 @@ Conn_Handler(void)
 		/* Look for non-empty read buffers ... */
 		for (i = 0; i < Pool_Size; i++) {
 			if ((My_Connections[i].sock > NONE)
-			    && (array_bytes(&My_Connections[i].rbuf) > 0)
-			    && (My_Connections[i].delaytime <= t)) {
+			    && (array_bytes(&My_Connections[i].rbuf) > 0)) {
 				/* ... and try to handle the received data */
 				bytes_processed = Handle_Buffer(i);
 				/* if we processed data, and there might be
@@ -820,7 +924,7 @@ Conn_Handler(void)
 		 * which is the granularity with witch we handle "penalty
 		 * times" for example.
 		 * Note: tv_sec/usec are undefined(!) after io_dispatch()
-		 * returns, so we have to set it beforce each call to it! */
+		 * returns, so we have to set it before each call to it! */
 		tv.tv_usec = 0;
 		tv.tv_sec = 1;
 
@@ -833,6 +937,15 @@ Conn_Handler(void)
 			    PACKAGE_NAME);
 			exit(1);
 		}
+
+		/* Should ngIRCd timeout when idle? */
+		if (Conf_IdleTimeout > 0 && NumConnectionsAccepted > 0
+		    && idle_t > 0 && time(NULL) - idle_t >= Conf_IdleTimeout) {
+			LogDebug("Server idle timeout reached: %d second%s. Initiating shutdown ...",
+				 Conf_IdleTimeout,
+				 Conf_IdleTimeout == 1 ? "" : "s");
+			NGIRCd_SignalQuit = true;
+		}
 	}
 
 	if (NGIRCd_SignalQuit)
@@ -865,6 +978,9 @@ va_dcl
 #endif
 {
 	char buffer[COMMAND_LEN];
+#ifdef ICONV
+	char *ptr, *message;
+#endif
 	size_t len;
 	bool ok;
 	va_list ap;
@@ -893,7 +1009,7 @@ va_dcl
 		 *
 		 * So we have a big problem here: we should send more bytes
 		 * to the network than we are allowed to and we don't know
-		 * the originator (any more). The "old" behaviour of blaming
+		 * the originator (any more). The "old" behavior of blaming
 		 * the receiver ("next hop") is a bad idea (it could be just
 		 * an other server only routing the message!), so the only
 		 * option left is to shorten the string and to hope that the
@@ -905,6 +1021,16 @@ va_dcl
 			CUT_TXTSUFFIX);
 	}
 
+#ifdef ICONV
+	ptr = strchr(buffer + 1, ':');
+	if (ptr) {
+		ptr++;
+		message = Conn_EncodingTo(Idx, ptr);
+		if (message != ptr)
+			strlcpy(ptr, message, sizeof(buffer) - (ptr - buffer));
+	}
+#endif
+
 #ifdef SNIFFER
 	if (NGIRCd_Sniffer)
 		Log(LOG_DEBUG, " -> connection %d: '%s'.", Idx, buffer);
@@ -918,6 +1044,30 @@ va_dcl
 	return ok;
 } /* Conn_WriteStr */
 
+GLOBAL char*
+Conn_Password( CONN_ID Idx )
+{
+	assert( Idx > NONE );
+	if (My_Connections[Idx].pwd == NULL)
+		return (char*)"\0";
+	else
+		return My_Connections[Idx].pwd;
+} /* Conn_Password */
+
+GLOBAL void
+Conn_SetPassword( CONN_ID Idx, const char *Pwd )
+{
+	assert( Idx > NONE );
+
+	if (My_Connections[Idx].pwd)
+		free(My_Connections[Idx].pwd);
+
+	My_Connections[Idx].pwd = strdup(Pwd);
+	if (My_Connections[Idx].pwd == NULL) {
+		Log(LOG_EMERG, "Can't allocate memory! [Conn_SetPassword]");
+		exit(1);
+	}
+} /* Conn_SetPassword */
 
 /**
  * Append Data to the outbound write buffer of a connection.
@@ -1033,8 +1183,8 @@ Conn_Close( CONN_ID Idx, const char *LogMsg, const char *FwdMsg, bool InformClie
 	/* Is this link already shutting down? */
 	if( Conn_OPTION_ISSET( &My_Connections[Idx], CONN_ISCLOSING )) {
 		/* Conn_Close() has been called recursively for this link;
-		 * probabe reason: Handle_Write() failed  -- see below. */
-		LogDebug("Recursive request to close connection: %d", Idx );
+		 * probable reason: Handle_Write() failed -- see below. */
+		LogDebug("Recursive request to close connection %d!", Idx );
 		return;
 	}
 
@@ -1044,7 +1194,7 @@ Conn_Close( CONN_ID Idx, const char *LogMsg, const char *FwdMsg, bool InformClie
 	Conn_OPTION_ADD( &My_Connections[Idx], CONN_ISCLOSING );
 
 	port = ng_ipaddr_getport(&My_Connections[Idx].addr);
-	Log(LOG_INFO, "Shutting down connection %d (%s) with %s:%d ...", Idx,
+	Log(LOG_INFO, "Shutting down connection %d (%s) with \"%s:%d\" ...", Idx,
 	    LogMsg ? LogMsg : FwdMsg, My_Connections[Idx].host, port);
 
 	/* Search client, if any */
@@ -1080,7 +1230,7 @@ Conn_Close( CONN_ID Idx, const char *LogMsg, const char *FwdMsg, bool InformClie
 	c = Conn_GetClient( Idx );
 #ifdef SSL_SUPPORT
 	if ( Conn_OPTION_ISSET( &My_Connections[Idx], CONN_SSL )) {
-		Log(LOG_INFO, "SSL connection %d shutting down ...", Idx);
+		LogDebug("SSL connection %d shutting down ...", Idx);
 		ConnSSL_Free(&My_Connections[Idx]);
 	}
 #endif
@@ -1118,7 +1268,7 @@ Conn_Close( CONN_ID Idx, const char *LogMsg, const char *FwdMsg, bool InformClie
 		in_p = (int)(( in_k * 100 ) / in_z_k );
 		out_p = (int)(( out_k * 100 ) / out_z_k );
 		Log(LOG_INFO,
-		    "Connection %d with %s:%d closed (in: %.1fk/%.1fk/%d%%, out: %.1fk/%.1fk/%d%%).",
+		    "Connection %d with \"%s:%d\" closed (in: %.1fk/%.1fk/%d%%, out: %.1fk/%.1fk/%d%%).",
 		    Idx, My_Connections[Idx].host, port,
 		    in_k, in_z_k, in_p, out_k, out_z_k, out_p);
 	}
@@ -1126,7 +1276,7 @@ Conn_Close( CONN_ID Idx, const char *LogMsg, const char *FwdMsg, bool InformClie
 #endif
 	{
 		Log(LOG_INFO,
-		    "Connection %d with %s:%d closed (in: %.1fk, out: %.1fk).",
+		    "Connection %d with \"%s:%d\" closed (in: %.1fk, out: %.1fk).",
 		    Idx, My_Connections[Idx].host, port,
 		    in_k, out_k);
 	}
@@ -1146,6 +1296,8 @@ Conn_Close( CONN_ID Idx, const char *LogMsg, const char *FwdMsg, bool InformClie
 
 	array_free(&My_Connections[Idx].rbuf);
 	array_free(&My_Connections[Idx].wbuf);
+	if (My_Connections[Idx].pwd != NULL)
+		free(My_Connections[Idx].pwd);
 
 	/* Clean up connection structure (=free it) */
 	Init_Conn_Struct( Idx );
@@ -1155,6 +1307,8 @@ Conn_Close( CONN_ID Idx, const char *LogMsg, const char *FwdMsg, bool InformClie
 		NumConnections--;
 	LogDebug("Shutdown of connection %d completed, %ld connection%s left.",
 		 Idx, NumConnections, NumConnections != 1 ? "s" : "");
+
+	idle_t = NumConnections > 0 ? 0 : time(NULL);
 } /* Conn_Close */
 
 
@@ -1298,9 +1452,14 @@ Handle_Write( CONN_ID Idx )
 		if (errno == EAGAIN || errno == EINTR)
 			return true;
 
-		Log(LOG_ERR, "Write error on connection %d (socket %d): %s!",
-		    Idx, My_Connections[Idx].sock, strerror(errno));
-		Conn_Close(Idx, "Write error!", NULL, false);
+		if (!Conn_OPTION_ISSET(&My_Connections[Idx], CONN_ISCLOSING))
+			Log(LOG_ERR,
+			    "Write error on connection %d (socket %d): %s!",
+			    Idx, My_Connections[Idx].sock, strerror(errno));
+		else
+			LogDebug("Recursive write error on connection %d (socket %d): %s!",
+				 Idx, My_Connections[Idx].sock, strerror(errno));
+		Conn_Close(Idx, "Write error", NULL, false);
 		return false;
 	}
 
@@ -1336,17 +1495,18 @@ Count_Connections(ng_ipaddr_t *a)
  * Initialize new client connection on a listening socket.
  *
  * @param Sock	Listening socket descriptor.
+ * @param IsSSL	true if this socket expects SSL-encrypted data.
  * @returns	Accepted socket descriptor or -1 on error.
  */
 static int
-New_Connection(int Sock)
+New_Connection(int Sock, UNUSED bool IsSSL)
 {
 #ifdef TCPWRAP
 	struct request_info req;
 #endif
 	ng_ipaddr_t new_addr;
 	char ip_str[NG_INET_ADDRSTRLEN];
-	int new_sock, new_sock_len, identsock;
+	int new_sock, new_sock_len;
 	CLIENT *c;
 	long cnt;
 
@@ -1463,35 +1623,65 @@ New_Connection(int Sock)
 
 	Client_SetHostname(c, My_Connections[new_sock].host);
 
-	Log(LOG_INFO, "Accepted connection %d from %s:%d on socket %d.",
+	Log(LOG_INFO, "Accepted connection %d from \"%s:%d\" on socket %d.",
 	    new_sock, My_Connections[new_sock].host,
 	    ng_ipaddr_getport(&new_addr), Sock);
-
-	identsock = new_sock;
-#ifdef IDENTAUTH
-	if (!Conf_Ident)
-		identsock = -1;
-#endif
-	if (Conf_DNS) {
-		if (Conf_NoticeAuth) {
-#ifdef IDENTAUTH
-			if (Conf_Ident)
-				(void)Conn_WriteStr(new_sock,
-					"NOTICE AUTH :*** Looking up your hostname and checking ident");
-			else
-#endif
-				(void)Conn_WriteStr(new_sock,
-					"NOTICE AUTH :*** Looking up your hostname");
-		}
-		Resolve_Addr(&My_Connections[new_sock].proc_stat, &new_addr,
-			     identsock, cb_Read_Resolver_Result);
-	}
-
 	Account_Connection();
+
+#ifdef SSL_SUPPORT
+	/* Delay connection initalization until SSL handshake is finished */
+	if (!IsSSL)
+#endif
+		Conn_StartLogin(new_sock);
+
 	return new_sock;
 } /* New_Connection */
 
 
+/**
+ * Finish connection initialization, start resolver subprocess.
+ *
+ * @param Idx Connection index.
+ */
+GLOBAL void
+Conn_StartLogin(CONN_ID Idx)
+{
+	int ident_sock = -1;
+
+	assert(Idx >= 0);
+
+	/* Nothing to do if DNS (and resolver subprocess) is disabled */
+	if (!Conf_DNS)
+		return;
+
+#ifdef IDENTAUTH
+	/* Should we make an IDENT request? */
+	if (Conf_Ident)
+		ident_sock = My_Connections[Idx].sock;
+#endif
+
+	if (Conf_NoticeAuth) {
+		/* Send "NOTICE AUTH" messages to the client */
+#ifdef IDENTAUTH
+		if (Conf_Ident)
+			(void)Conn_WriteStr(Idx,
+				"NOTICE AUTH :*** Looking up your hostname and checking ident");
+		else
+#endif
+			(void)Conn_WriteStr(Idx,
+				"NOTICE AUTH :*** Looking up your hostname");
+		/* Send buffered data to the client, but break on errors
+		 * because Handle_Write() would have closed the connection
+		 * again in this case! */
+		if (!Handle_Write(Idx))
+			return;
+	}
+
+	Resolve_Addr(&My_Connections[Idx].proc_stat, &My_Connections[Idx].addr,
+		     ident_sock, cb_Read_Resolver_Result);
+}
+
+
 /**
  * Update global connection counters.
  */
@@ -1499,6 +1689,7 @@ static void
 Account_Connection(void)
 {
 	NumConnections++;
+	idle_t = 0;
 	if (NumConnections > NumConnectionsMax)
 		NumConnectionsMax = NumConnections;
 	LogDebug("Total number of connections now %lu (max %lu).",
@@ -1528,7 +1719,7 @@ Socket2Index( int Sock )
 
 
 /**
- * Read data from the network to the read buffer. If an error occures,
+ * Read data from the network to the read buffer. If an error occurs,
  * the socket of this connection will be shut down.
  *
  * @param Idx	Connection index.
@@ -1566,13 +1757,10 @@ Read_Request( CONN_ID Idx )
 #endif
 	len = read(My_Connections[Idx].sock, readbuf, sizeof(readbuf));
 	if (len == 0) {
-		Log(LOG_INFO, "%s:%u (%s) is closing the connection ...",
-				My_Connections[Idx].host,
-				(unsigned int) ng_ipaddr_getport(&My_Connections[Idx].addr),
-				ng_ipaddr_tostr(&My_Connections[Idx].addr));
-		Conn_Close(Idx,
-			   "Socket closed!", "Client closed connection",
-			   false);
+		LogDebug("Client \"%s:%u\" is closing connection %d ...",
+			 My_Connections[Idx].host,
+			 ng_ipaddr_tostr(&My_Connections[Idx].addr), Idx);
+		Conn_Close(Idx, NULL, "Client closed connection", false);
 		return;
 	}
 
@@ -1580,7 +1768,7 @@ Read_Request( CONN_ID Idx )
 		if( errno == EAGAIN ) return;
 		Log(LOG_ERR, "Read error on connection %d (socket %d): %s!",
 		    Idx, My_Connections[Idx].sock, strerror(errno));
-		Conn_Close(Idx, "Read error!", "Client closed connection",
+		Conn_Close(Idx, "Read error", "Client closed connection",
 			   false);
 		return;
 	}
@@ -1680,7 +1868,7 @@ Handle_Buffer(CONN_ID Idx)
 		maxcmd = (int)(Client_UserCount() / 5)
 		       + MAX_COMMANDS_SERVER_MIN;
 		/* Allow servers to handle even more commands while peering
-		 * to speed up server login and network synchronisation. */
+		 * to speed up server login and network synchronization. */
 		if (Conn_LastPing(Idx) == 0)
 			maxcmd *= 5;
 		break;
@@ -1839,10 +2027,10 @@ Check_Connections(void)
 				if (My_Connections[i].lastping <
 				    time(NULL) - Conf_PongTimeout) {
 					/* Timeout */
-					LogDebug
-					    ("Connection %d: Ping timeout: %d seconds.",
-					     i, Conf_PongTimeout);
-					snprintf(msg, sizeof(msg), "Ping timeout: %d seconds", Conf_PongTimeout);
+					snprintf(msg, sizeof(msg),
+						 "Ping timeout: %d seconds",
+						 Conf_PongTimeout);
+					LogDebug("Connection %d: %s.", i, msg);
 					Conn_Close(i, NULL, msg, true);
 				}
 			} else if (My_Connections[i].lastdata <
@@ -1945,6 +2133,7 @@ New_Server( int Server , ng_ipaddr_t *dest)
 
 	if (!ng_ipaddr_tostr_r(dest, ip_str)) {
 		Log(LOG_WARNING, "New_Server: Could not convert IP to string");
+		Conf_Server[Server].conn_id = NONE;
 		return;
 	}
 
@@ -1959,11 +2148,14 @@ New_Server( int Server , ng_ipaddr_t *dest)
 	if (new_sock < 0) {
 		Log(LOG_CRIT, "Can't create socket (af %d): %s!",
 		    af_dest, strerror(errno));
+		Conf_Server[Server].conn_id = NONE;
 		return;
 	}
 
-	if (!Init_Socket(new_sock))
+	if (!Init_Socket(new_sock)) {
+		Conf_Server[Server].conn_id = NONE;
 		return;
+	}
 
 	/* is a bind address configured? */
 	res = ng_ipaddr_af(&Conf_Server[Server].bind_addr);
@@ -1979,6 +2171,7 @@ New_Server( int Server , ng_ipaddr_t *dest)
 	if(( res != 0 ) && ( errno != EINPROGRESS )) {
 		Log( LOG_CRIT, "Can't connect socket: %s!", strerror( errno ));
 		close( new_sock );
+		Conf_Server[Server].conn_id = NONE;
 		return;
 	}
 
@@ -1987,12 +2180,14 @@ New_Server( int Server , ng_ipaddr_t *dest)
 		    "Cannot allocate memory for server connection (socket %d)",
 		    new_sock);
 		close( new_sock );
+		Conf_Server[Server].conn_id = NONE;
 		return;
 	}
 
 	if (!io_event_create( new_sock, IO_WANTWRITE, cb_connserver)) {
 		Log(LOG_ALERT, "io_event_create(): could not add fd %d", strerror(errno));
 		close(new_sock);
+		Conf_Server[Server].conn_id = NONE;
 		return;
 	}
 
@@ -2007,6 +2202,7 @@ New_Server( int Server , ng_ipaddr_t *dest)
 	if (!c) {
 		Log( LOG_ALERT, "Can't establish connection: can't create client structure!" );
 		io_close(new_sock);
+		Conf_Server[Server].conn_id = NONE;
 		return;
 	}
 
@@ -2016,7 +2212,8 @@ New_Server( int Server , ng_ipaddr_t *dest)
 	Client_SetToken( c, TOKEN_OUTBOUND );
 
 	/* Register connection */
-	Conf_SetServer(Server, new_sock);
+	if (!Conf_SetServer(Server, new_sock))
+		return;
 	My_Connections[new_sock].sock = new_sock;
 	My_Connections[new_sock].addr = *dest;
 	My_Connections[new_sock].client = c;
@@ -2056,6 +2253,11 @@ Init_Conn_Struct(CONN_ID Idx)
 	My_Connections[Idx].lastdata = now;
 	My_Connections[Idx].lastprivmsg = now;
 	Proc_InitStruct(&My_Connections[Idx].proc_stat);
+
+#ifdef ICONV
+	My_Connections[Idx].iconv_from = (iconv_t)(-1);
+	My_Connections[Idx].iconv_to = (iconv_t)(-1);
+#endif
 } /* Init_Conn_Struct */
 
 
@@ -2168,7 +2370,7 @@ cb_Connect_to_Server(int fd, UNUSED short events)
 
 /**
  * Read results of a resolver sub-process from the pipe and update the
- * apropriate connection/client structure(s): hostname and/or IDENT user name.
+ * appropriate connection/client structure(s): hostname and/or IDENT user name.
  *
  * @param r_fd		File descriptor of the pipe to the sub-process.
  * @param events	(ignored IO specification)
@@ -2231,7 +2433,8 @@ cb_Read_Resolver_Result( int r_fd, UNUSED short events )
 		Client_SetHostname(c, readbuf);
 		if (Conf_NoticeAuth)
 			(void)Conn_WriteStr(i,
-					"NOTICE AUTH :*** Found your hostname");
+					"NOTICE AUTH :*** Found your hostname: %s",
+					My_Connections[i].host);
 #ifdef IDENTAUTH
 		++identptr;
 		if (*identptr) {
@@ -2256,20 +2459,32 @@ cb_Read_Resolver_Result( int r_fd, UNUSED short events )
 			}
 			if (Conf_NoticeAuth) {
 				(void)Conn_WriteStr(i,
-					"NOTICE AUTH :*** Got %sident response",
-					*ptr ? "invalid " : "");
+					"NOTICE AUTH :*** Got %sident response%s%s",
+					*ptr ? "invalid " : "",
+					*ptr ? "" : ": ",
+					*ptr ? "" : identptr);
 			}
-		} else {
+		} else if(Conf_Ident) {
 			Log(LOG_INFO, "IDENT lookup for connection %d: no result.", i);
-			if (Conf_NoticeAuth && Conf_Ident)
+			if (Conf_NoticeAuth)
 				(void)Conn_WriteStr(i,
 					"NOTICE AUTH :*** No ident response");
 		}
 #endif
+
+		if (Conf_NoticeAuth) {
+			/* Send buffered data to the client, but break on
+			 * errors because Handle_Write() would have closed
+			 * the connection again in this case! */
+			if (!Handle_Write(i))
+				return;
+		}
+
 		Class_HandleServerBans(c);
 	}
 #ifdef DEBUG
-		else Log( LOG_DEBUG, "Resolver: discarding result for already registered connection %d.", i );
+	else
+		LogDebug("Resolver: discarding result for already registered connection %d.", i);
 #endif
 } /* cb_Read_Resolver_Result */
 
@@ -2388,7 +2603,7 @@ Conn_SetAuthPing(CONN_ID Idx, long ID)
 #ifdef SSL_SUPPORT
 
 /**
- * Get information about used SSL chiper.
+ * Get information about used SSL cipher.
  *
  * @param Idx	Connection index number.
  * @param buf	Buffer for returned information text.
@@ -2420,6 +2635,45 @@ Conn_UsesSSL(CONN_ID Idx)
 	return Conn_OPTION_ISSET(&My_Connections[Idx], CONN_SSL);
 }
 
+
+GLOBAL char *
+Conn_GetCertFp(CONN_ID Idx)
+{
+	if (Idx < 0)
+		return NULL;
+	assert(Idx < (int) array_length(&My_ConnArray, sizeof(CONNECTION)));
+	return ConnSSL_GetCertFp(&My_Connections[Idx]);
+}
+
+
+GLOBAL bool
+Conn_SetCertFp(CONN_ID Idx, const char *fingerprint)
+{
+	if (Idx < 0)
+		return false;
+	assert(Idx < (int) array_length(&My_ConnArray, sizeof(CONNECTION)));
+	return ConnSSL_SetCertFp(&My_Connections[Idx], fingerprint);
+}
+#else
+GLOBAL bool
+Conn_UsesSSL(UNUSED CONN_ID Idx)
+{
+	return false;
+}
+
+
+GLOBAL char *
+Conn_GetCertFp(UNUSED CONN_ID Idx)
+{
+	return NULL;
+}
+
+
+GLOBAL bool
+Conn_SetCertFp(UNUSED CONN_ID Idx, UNUSED const char *fingerprint)
+{
+	return true;
+}
 #endif
 
 
